"""Async classes and exceptions for Simple-Salesforce API"""

import aiofiles
import asyncio
import base64
from collections import OrderedDict
from datetime import datetime
from functools import partial, wraps
import json
import logging
from pathlib import Path
import re
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    TypeVar,
    Tuple,
)
from urllib.parse import urljoin, urlparse
import warnings

import httpx
from typing_extensions import ParamSpec

from simple_salesforce.api import DEFAULT_API_VERSION, PerAppUsage, Usage
from simple_salesforce.exceptions import (
    SalesforceExpiredSession,
    SalesforceGeneralError,
)
from simple_salesforce.util import date_to_iso8601, Headers, Proxies
from .bulk import AsyncSFBulkHandler
from .bulk2 import AsyncSFBulk2Handler
from .login import AsyncSalesforceLogin
from .metadata import AsyncSfdcMetadataApi
from .aio_util import create_session_factory, exception_handler


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


async def build_async_salesforce_client(
    username: Optional[str] = None,
    password: Optional[str] = None,
    security_token: Optional[str] = None,
    session_id: Optional[str] = None,
    instance: Optional[str] = None,
    instance_url: Optional[str] = None,
    organizationId: Optional[str] = None,
    version: str = DEFAULT_API_VERSION,
    proxies: Proxies | None = None,
    client_id: Optional[str] = None,
    domain: Optional[str] = None,
    consumer_key: Optional[str] = None,
    consumer_secret: Optional[str] = None,
    privatekey_file: Optional[str] = None,
    privatekey: Optional[str] = None,
    parse_float: Optional[Callable[[str], Any]] = None,
    object_pairs_hook: Optional[Callable[[List[Tuple[Any, Any]]], Any]] = OrderedDict,
    request_timeout_seconds: Optional[int] = None,
) -> "AsyncSalesforce":
    """
    Reasons for this builder function:

        - Async code is not allowed in __init__ method.
        - Also, the original Salesforce client did not store
        username/password on the instance.

    Thus, this helper functions constructs an instance
    and issues an *async* login call without storing auth parameters
    in memory. This function has same kwargs as the original
    `Salesforce` client class.

    Available kwargs
    Password Authentication:
    * username -- the Salesforce username to use for authentication
    * password -- the password for the username
    * security_token -- the security token for the username
    * domain -- The domain to using for connecting to Salesforce. Use
                common domains, such as 'login' or 'test', or
                Salesforce My domain. If not used, will default to
                'login'.

    OAuth 2.0 Connected App Token Authentication:
    * consumer_key -- the consumer key generated for the user
    * consumer_secret -- the consumer secret generated for the user

    OAuth 2.0 JWT Bearer Token Authentication:
    * consumer_key -- the consumer key generated for the user

    Then either
    * privatekey_file -- the path to the private key file used
                            for signing the JWT token
    OR
    * privatekey -- the private key to use
                        for signing the JWT token

    Direct Session and Instance Access:

    * session_id -- Access token for this session

    Then either
    * instance -- Domain of your Salesforce instance, i.e.
        `na1.salesforce.com`
    OR
    * instance_url -- Full URL of your instance i.e.
        `https://na1.salesforce.com

    Universal Kwargs:
    * version -- the version of the Salesforce API to use, for example
                    `29.0`
    * proxies -- the optional map of scheme to proxy server (note: httpx style!)
    * parse_float -- Function to parse float values with. Is passed along to
                     https://docs.python.org/3/library/json.html#json.load
    * object_pairs_hook -- Function to parse ordered list of pairs in json.
                           To use python 'dict' change it to None or dict.
    * request_timeout_seconds -- Optional request-timeout setting in seconds.
    """
    session_factory = create_session_factory(proxies, timeout=request_timeout_seconds)

    if domain is None:
        domain = "login"

    instance_kwargs = {
        "version": version,
        "proxies": proxies,
        "domain": domain,
        "parse_float": parse_float,
        "object_pairs_hook": object_pairs_hook,
        "request_timeout_seconds": request_timeout_seconds,
        "session_factory": session_factory,
    }

    # Determine if the user wants to use our username/password auth or pass
    # in their own information
    login_refresh = None
    if all(arg is not None for arg in (username, password, security_token)):
        instance_kwargs["auth_type"] = "password"

        # Pass along the username/password to our login helper
        login_refresh = partial(
            AsyncSalesforceLogin,
            session_factory=session_factory,
            username=username,
            password=password,
            security_token=security_token,
            sf_version=version,
            proxies=proxies,
            client_id=client_id,
            domain=domain,
        )
    elif all(arg is not None for arg in (session_id, instance or instance_url)):
        instance_kwargs["auth_type"] = "direct"
        instance_kwargs["session_id"] = session_id

        # If the user provides the full url (as returned by the OAuth
        # interface for example) extract the hostname (which we rely on)
        if instance_url is not None:
            instance_kwargs["sf_instance"] = urlparse(instance_url).hostname
            port = urlparse(instance_url).port
            if port not in (None, 443):
                instance_kwargs["sf_instance"] += f":{port}"  # type: ignore
        else:
            instance_kwargs["sf_instance"] = instance

    elif all(arg is not None for arg in (username, password, organizationId)):
        instance_kwargs["auth_type"] = "ipfilter"
        login_refresh = partial(
            AsyncSalesforceLogin,
            session_factory=session_factory,
            username=username,
            password=password,
            organizationId=organizationId,
            sf_version=version,
            proxies=proxies,
            client_id=client_id,
            domain=domain,
        )
    elif all(
        arg is not None for arg in (username, password, consumer_key, consumer_secret)
    ):
        instance_kwargs["auth_type"] = "password"

        # Pass along the username/password to our login helper
        login_refresh = partial(
            AsyncSalesforceLogin,
            session_factory=session_factory,
            username=username,
            password=password,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            proxies=proxies,
            domain=domain,
        )
    elif all(
        arg is not None
        for arg in (username, consumer_key, privatekey_file or privatekey)
    ):
        instance_kwargs["auth_type"] = "jwt-bearer"

        # Pass along the username/password to our login helper
        login_refresh = partial(
            AsyncSalesforceLogin,
            session_factory=session_factory,
            username=username,
            instance_url=instance_url,
            consumer_key=consumer_key,
            privatekey_file=privatekey_file,
            privatekey=privatekey,
            proxies=proxies,
            domain=domain,
        )
    elif all(arg is not None for arg in (consumer_key, consumer_secret, domain)):
        instance_kwargs["auth_type"] = "client-credentials"
        login_refresh = partial(
            AsyncSalesforceLogin,
            session_factory=session_factory,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            proxies=proxies,
            domain=domain,
        )
    else:
        raise TypeError("You must provide login information or an instance and token")

    if login_refresh is not None:
        session_id, sf_instance = await login_refresh()
        instance_kwargs["session_id"] = session_id
        instance_kwargs["sf_instance"] = sf_instance
        instance_kwargs["login_refresh"] = login_refresh

    # TODO: type this dict so we can tell if we're calling constructor properly
    return AsyncSalesforce(**instance_kwargs)  # type: ignore


# See https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
Param = ParamSpec("Param")
RetType = TypeVar("RetType")


# This decorator is used to retry a function ONCE if the session has expired
def async_retry_expired_session(
    async_fn: Callable[Param, Awaitable[RetType]]
) -> Callable[Param, Awaitable[RetType]]:
    """Decorator for AsyncSalesforce methods with retries for expired session"""
    if not asyncio.iscoroutinefunction(async_fn):
        raise ValueError(
            (
                "async_retry_expired_session can only decorate async "
                f"functions. {async_fn.__name__}"
            )
        )

    @wraps(async_fn)
    async def inner(client: Any, *args: Param.args, **kwargs: Param.kwargs) -> RetType:
        try:
            return await async_fn(client, *args, **kwargs)
        except SalesforceExpiredSession:
            if not client.can_refresh:
                # No automatic refresh
                raise

            await client.refresh_session()
            return await async_fn(client, *args, **kwargs)

    # Mypy doesn't like the _Wrapped[...] type returned by wraps
    # see: https://github.com/python/mypy/issues/17166
    return inner  # type: ignore


# pylint: disable=too-many-instance-attributes
class AsyncSalesforce:
    """Salesforce Instance

    An instance of AsyncSalesforce is a handy way to wrap a Salesforce session
    for easy use of the Salesforce REST API. All http network calls are async.
    """

    _parse_float = None
    _object_pairs_hook = OrderedDict

    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
    def __init__(
        self,
        version: str = DEFAULT_API_VERSION,
        proxies: Proxies | None = None,
        session_factory: Optional[Callable[[], httpx.AsyncClient]] = None,
        domain: Optional[str] = None,
        auth_type: Optional[str] = None,
        session_id: Optional[str] = None,
        sf_instance: Optional[str] = None,
        parse_float: Optional[Callable[[str], Any]] = None,
        object_pairs_hook: Optional[
            Callable[[List[Tuple[Any, Any]]], Any]
        ] = OrderedDict,
        request_timeout_seconds: Optional[int] = None,
        login_refresh: Optional[Callable[[], Awaitable[Tuple[str, str]]]] = None,
    ):
        """Initialize the instance with the given parameters.

        Universal Kwargs:

        * domain -- The domain to using for connecting to Salesforce. Use
                    common domains, such as 'login' or 'test', or
                    Salesforce My domain. If not used, will default to
                    'login'.
        * version -- the version of the Salesforce API to use, for example
                     `29.0`
        * proxies -- the optional map of scheme to proxy server (httpx style)
        * session_factory -- Method for generating a new httpx session (AsyncClient).
                    This enables the use of httpx AsyncClient features not
                    otherwise exposed by simple_salesforce.
        * parse_float -- Function to parse float values with. Is passed along to
                         https://docs.python.org/3/library/json.html#json.load
        * object_pairs_hook -- Function to parse ordered list of pairs in json.
                               To use python 'dict' change it to None or dict.
        * request_timeout_seconds -- Optional request-timeout setting in seconds
                                     (for each request; for more timeout options, use `session`).
        * login_refresh -- A function can be supplied to *run* logins (refreshes) in the future.
                           It must return a tuple of `(session_id, sf_instance_name)`
        """
        # Determine if the user passed in the optional version and/or
        # domain kwargs
        self.auth_type = auth_type
        self.session_id: str | None = session_id
        self.sf_instance = sf_instance
        self.sf_version = version
        self.domain = domain or ""
        self._proxies = proxies
        self.request_timeout_seconds = request_timeout_seconds
        self.login_refresh = login_refresh

        # override custom session proxies dance
        if not session_factory:
            session_factory = create_session_factory(
                self._proxies, timeout=self.request_timeout_seconds
            )
        self.session_factory = session_factory

        self.auth_site = f"https://{domain}.salesforce.com"
        self.headers = self._generate_headers()

        self.base_url = f"https://{self.sf_instance}/services/data/v{self.sf_version}/"
        self.apex_url = f"https://{self.sf_instance}/services/apexrest/"
        self.bulk_url = f"https://{self.sf_instance}/services/async/{self.sf_version}/"
        self.bulk2_url = (
            f"https://{self.sf_instance}/services/data/v{self.sf_version}/jobs/"
        )
        self.metadata_url = (
            f"https://{self.sf_instance}/services/Soap/m/{self.sf_version}/"
        )
        self.oauth2_url = f"https://{self.sf_instance}/services/oauth2/"
        self.tooling_url = f"{self.base_url}tooling/"

        self.api_usage: MutableMapping[str, Usage | PerAppUsage] = {}
        self._parse_float = parse_float
        self._object_pairs_hook = object_pairs_hook  # type: ignore[assignment]
        self._mdapi: AsyncSfdcMetadataApi | None = None

    @property
    def mdapi(self) -> AsyncSfdcMetadataApi:
        """Utility to interact with metadata api functionality"""
        if not self._mdapi:
            self._mdapi = AsyncSfdcMetadataApi(
                session_factory=self.session_factory,
                session_id=self.session_id,
                instance=self.sf_instance,
                metadata_url=self.metadata_url,
                api_version=self.sf_version,
                headers=self.headers,
            )
        return self._mdapi

    def _generate_headers(self) -> Dict[str, str]:
        """Utility to generate headers when refreshing the session"""
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.session_id or ''}",
            "X-PrettyPrint": "1",
        }
        return self.headers

    async def refresh_session(self) -> str:
        """Utility to refresh the session when expired"""
        if self.login_refresh is None:
            raise RuntimeError(
                "The simple_salesforce session can not refreshed if a "
                "session id has been provided."
            )
        self.session_id, self.sf_instance = await self.login_refresh()
        self._generate_headers()
        return self.session_id

    @property
    def can_refresh(self) -> bool:
        return self.login_refresh is not None

    async def describe(self, **kwargs: Any) -> Optional[Any]:
        """Describes all available objects

        Arguments:

        * keyword arguments supported by requests.request (e.g. json, timeout)
        """
        url = self.base_url + "sobjects"
        result = await self._call_salesforce("GET", url, name="describe", **kwargs)

        json_result = self.parse_result_to_json(result)
        if len(json_result) == 0:
            return None

        return json_result

    async def is_sandbox(self) -> Optional[bool]:
        """After connection returns is the organization in a sandbox"""
        is_sandbox = None
        if self.session_id:
            is_sandbox = await self.query_all(
                "SELECT IsSandbox FROM Organization LIMIT 1"
            )
            is_sandbox = is_sandbox.get("records", [{"IsSandbox": None}])[0].get(
                "IsSandbox"
            )
        return is_sandbox

    # SObject Handler
    def __getattr__(self, name: str) -> Any:
        """Returns an `AsyncSFType` instance for the given Salesforce object type
        (given in `name`).

        The magic part of the SalesforceAPI, this function translates
        calls such as `salesforce_api_instance.Lead.metadata()` into fully
        constituted `AsyncSFType` instances to make a nice Python API wrapper
        for the REST API.

        Arguments:

        * name -- the name of a Salesforce object type, e.g. Lead or Contact
        """

        # fix to enable serialization
        # (https://github.com/heroku/simple-salesforce/issues/60)
        if name.startswith("__"):
            return super().__getattr__(name)

        if name == "bulk":
            # Deal with bulk API functions
            return AsyncSFBulkHandler(
                self.session_id,
                self.bulk_url,
                self._proxies,
                session_factory=self.session_factory,
            )
        if name == "bulk2":
            return AsyncSFBulk2Handler(
                self.session_id,
                self.bulk2_url,
                self.proxies,
                session_factory=self.session_factory,
            )

        return AsyncSFType(
            name,
            self.session_id or "",
            self.sf_instance or "",
            sf_version=self.sf_version,
            proxies=self._proxies,
            salesforce=self,
            session_factory=self.session_factory,
            object_pairs_hook=self._object_pairs_hook,
            parse_float=self._parse_float,
        )

    # User utility methods
    async def set_password(self, user: str, password: str) -> Optional[Any]:
        """Sets the password of a user

        salesforce dev documentation link:
        https://www.salesforce.com/us/developer/docs/api_rest/Content
        /dome_sobject_user_password.htm

        Arguments:

        * user: the userID of the user to set
        * password: the new password
        """
        url = f"{self.base_url}sobjects/User/{user}/password"
        params = {"NewPassword": password}

        result = await self._call_salesforce("POST", url, content=json.dumps(params))

        if result.status_code == 204:
            return None

        # salesforce return 204 No Content when the request is successful
        if result.status_code != 200:
            raise SalesforceGeneralError(
                url, result.status_code, "User", result.content
            )
        return self.parse_result_to_json(result)

    # Generic Rest Function
    async def restful(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        method: str = "GET",
        **kwargs: Any,
    ) -> Optional[Any]:
        """Allows you to make a direct REST call if you know the path

        Arguments:

        * path: The path of the request
            Example: sobjects/User/ABC123/password'
        * params: dict of parameters to pass to the path
        * method: HTTP request method, default GET
        * other arguments supported by requests.request (e.g. json, timeout)
        """

        url = self.base_url + path
        result = await self._call_salesforce(
            method, url, name=path, params=params, **kwargs
        )
        json_result = self.parse_result_to_json(result)

        if len(json_result) == 0:
            return None

        return json_result

    # OAuth Endpoints Function
    async def oauth2(
        self, path: str, params: Optional[Dict[str, Any]] = None, method: str = "GET"
    ) -> Optional[Any]:
        """Allows you to make a request to OAuth endpoints if you know the path

        Arguments:

        * path: The path of the request
            Example: /services/oauth2/token'
        * params: dict of parameters to pass to the path
        * method: HTTP request method, default GET
        * other arguments supported by requests.request (e.g. json, timeout)
        """
        url = self.oauth2_url + path
        result = await self._call_salesforce(method, url, name=path, params=params)

        content_type = result.headers.get("Content-Type")
        json_result = (
            self.parse_result_to_json(result)
            if content_type is not None and "json" in content_type
            else None
        )

        return None if json_result and len(json_result) == 0 else json_result

    # Search Functions
    async def search(self, search: str) -> Any:
        """Returns the result of a Salesforce search as a dict decoded from
        the Salesforce response JSON payload.

        Arguments:

        * search -- the fully formatted SOSL search string, e.g.
                    `FIND {Waldo}`
        """
        url = self.base_url + "search/"

        # `requests` will correctly encode the query string passed as `params`
        params = {"q": search}
        result = await self._call_salesforce("GET", url, name="search", params=params)

        json_result = self.parse_result_to_json(result)
        if len(json_result) == 0:
            return None

        return json_result

    @async_retry_expired_session
    async def quick_search(self, search: str) -> Any:
        """Returns the result of a Salesforce search as a dict decoded from
        the Salesforce response JSON payload.

        Arguments:

        * search -- the non-SOSL search string, e.g. `Waldo`. This search
                    string will be wrapped to read `FIND {Waldo}` before being
                    sent to Salesforce
        """
        search_string = f"FIND {{{search}}}"
        return await self.search(search_string)

    async def limits(self, **kwargs: Any) -> Any:
        """Return the result of a Salesforce request to list Organization
        limits.
        """
        url = self.base_url + "limits/"
        result = await self._call_salesforce("GET", url, **kwargs)

        if result.status_code != 200:
            exception_handler(result)
        return self.parse_result_to_json(result)

    # Query Handler
    async def query(
        self, query: str, include_deleted: bool = False, **kwargs: Any
    ) -> Any:
        """Return the result of a Salesforce SOQL query as a dict decoded from
        the Salesforce response JSON payload.

        Arguments:

        * query -- the SOQL query to send to Salesforce, e.g.
                   SELECT Id FROM Lead WHERE Email = "waldo@somewhere.com"
        * include_deleted -- True if deleted records should be included
        """
        url = self.base_url + ("queryAll/" if include_deleted else "query/")
        params = {"q": query}
        # `requests` will correctly encode the query string passed as `params`
        result = await self._call_salesforce(
            "GET", url, name="query", params=params, **kwargs
        )
        return self.parse_result_to_json(result)

    async def query_more(
        self,
        next_records_identifier: str,
        identifier_is_url: bool = False,
        include_deleted: bool = False,
        **kwargs: Any,
    ) -> Any:
        """Retrieves more results from a query that returned more results
        than the batch maximum. Returns a dict decoded from the Salesforce
        response JSON payload.

        Arguments:

        * next_records_identifier -- either the Id of the next Salesforce
                                     object in the result, or a URL to the
                                     next record in the result.
        * identifier_is_url -- True if `next_records_identifier` should be
                               treated as a URL, False if
                               `next_records_identifier` should be treated as
                               an Id.
        * include_deleted -- True if the `next_records_identifier` refers to a
                             query that includes deleted records. Only used if
                             `identifier_is_url` is False
        """
        if identifier_is_url:
            # Don't use `self.base_url` here because the full URI is provided
            url = f"https://{self.sf_instance}{next_records_identifier}"
        else:
            endpoint = "queryAll" if include_deleted else "query"
            url = f"{self.base_url}{endpoint}/{next_records_identifier}"
        result = await self._call_salesforce("GET", url, name="query_more", **kwargs)

        return self.parse_result_to_json(result)

    async def query_all_iter(
        self, query: str, include_deleted: bool = False, **kwargs: Any
    ) -> AsyncGenerator[Any, None]:
        """This is an async-generator, lazy alternative to `query_all`:
        it does not construct the whole result set into one container,
        but returns objects from each page it retrieves from the API.

        Since `query_all` has always been eagerly executed, we reimplemented it
        using `query_all_iter`, only materializing the returned iterator to
        maintain backwards compatibility.

        The one big difference from `query_all` (apart from being lazy) is that
        we don't return a dictionary with `totalSize` and `done` here,
        we only return the records in an iterator.

        Arguments

        * query -- the SOQL query to send to Salesforce, e.g.
                   SELECT Id FROM Lead WHERE Email = "waldo@somewhere.com"
        * include_deleted -- True if the query should include deleted records.
        """

        result = await self.query(query, include_deleted=include_deleted, **kwargs)
        while True:
            for record in result["records"]:
                yield record
            # fetch next batch if we're not done else break out of loop
            if not result["done"]:
                result = await self.query_more(
                    result["nextRecordsUrl"], identifier_is_url=True, **kwargs
                )
            else:
                return

    @async_retry_expired_session
    async def query_all(
        self, query: str, include_deleted: bool = False, **kwargs: Any
    ) -> Dict[str, Any]:
        """Returns the full set of results for the `query`.
        This is a convenience wrapper around `query(...)` and
        `query_more(...)`. It calls `query_all_iter` and pages through
        the results internally.

        The returned dict is the decoded JSON payload from the final call to
        Salesforce, but with the `totalSize` field representing the full
        number of results retrieved and the `records` list representing the
        full list of records retrieved.

        Arguments

        * query -- the SOQL query to send to Salesforce, e.g.
                   SELECT Id FROM Lead WHERE Email = "waldo@somewhere.com"
        * include_deleted -- True if the query should include deleted records.
        """
        all_records = []
        response = self.query_all_iter(query, include_deleted=include_deleted, **kwargs)
        async for record in response:
            all_records.append(record)
        return {
            "records": all_records,
            "totalSize": len(all_records),
            "done": True,
        }

    async def toolingexecute(
        self,
        action: str,
        method: str = "GET",
        data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Makes an HTTP request to an TOOLING REST endpoint

        Arguments:

        * action -- The REST endpoint for the request.
        * method -- HTTP method for the request (default GET)
        * data -- A dict of parameters to send in a POST / PUT request
        * kwargs -- Additional kwargs to pass to `requests.request`
        """
        # If data is None, we should send an empty body, not "null", which is
        # None in json.
        json_data = json.dumps(data) if data is not None else None
        result = await self._call_salesforce(
            method,
            self.tooling_url + action,
            name="toolingexecute",
            content=json_data,
            **kwargs,
        )
        try:
            response_content = result.json()
        # pylint: disable=broad-except
        except Exception:
            response_content = result.text

        return response_content

    async def apexecute(
        self,
        action: str,
        method: str = "GET",
        data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Makes an HTTP request to an APEX REST endpoint

        Arguments:

        * action -- The REST endpoint for the request.
        * method -- HTTP method for the request (default GET)
        * data -- A dict of parameters to send in a POST / PUT request
        * kwargs -- Additional kwargs to pass to `requests.request`
        """
        # If data is None, we should send an empty body, not "null", which is
        # None in json.
        json_data = json.dumps(data) if data is not None else None
        result = await self._call_salesforce(
            method,
            self.apex_url + action,
            name="apexecute",
            content=json_data,
            **kwargs,
        )
        try:
            response_content = result.json()
        # pylint: disable=broad-except
        except Exception:
            response_content = result.text

        return response_content

    @async_retry_expired_session
    async def _call_salesforce(
        self, method: str, url: str, name: str = "", **kwargs: Any
    ) -> httpx.Response:
        """Utility method for performing HTTP call to Salesforce.

        Returns a `requests.result` object.
        """
        headers = self.headers.copy()
        additional_headers = kwargs.pop("headers", dict())
        headers.update(additional_headers)
        async with self.session_factory() as client:
            result = await client.request(method, url, headers=headers, **kwargs)

        if result.status_code >= 300:
            exception_handler(result, name=name)
        sforce_limit_info = result.headers.get("Sforce-Limit-Info")
        if sforce_limit_info:
            self.api_usage = self.parse_api_usage(sforce_limit_info)

        return result

    @staticmethod
    def parse_api_usage(
        sforce_limit_info: str,
    ) -> MutableMapping[str, Usage | PerAppUsage]:
        """parse API usage and limits out of the Sforce-Limit-Info header

        Arguments:

        * sforce_limit_info: The value of response header 'Sforce-Limit-Info'
            Example 1: 'api-usage=18/5000'
            Example 2: 'api-usage=25/5000;
                per-app-api-usage=17/250(appName=sample-connected-app)'
        """
        result = {}

        api_usage = re.match(
            r"[^-]?api-usage=(?P<used>\d+)/(?P<tot>\d+)", sforce_limit_info
        )
        pau = r".+per-app-api-usage=(?P<u>\d+)/(?P<t>\d+)\(appName=(?P<n>.+)\)"
        per_app_api_usage = re.match(pau, sforce_limit_info)

        if api_usage and api_usage.groups():
            groups = api_usage.groups()
            result["api-usage"] = Usage(used=int(groups[0]), total=int(groups[1]))
        if per_app_api_usage and per_app_api_usage.groups():
            groups = per_app_api_usage.groups()
            result["per-app-api-usage"] = PerAppUsage(
                used=int(groups[0]), total=int(groups[1]), name=groups[2]
            )

        return result

    # file-based deployment function
    async def deploy(
        self, zipfile: str | Path, sandbox: bool, **kwargs: Any
    ) -> Dict[str, Optional[str]]:
        """Deploy using the Salesforce Metadata API. Wrapper for
        SfdcMetaDataApi.deploy(...).

        Arguments:

        * zipfile: a .zip archive to deploy to an org, given as (
        "path/to/zipfile.zip")
        * options: salesforce DeployOptions in .json format.
            (https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta
            /api_meta/meta_deploy.htm)

        Returns a process id and state for this deployment.
        """
        asyncId, state = await self.mdapi.deploy(zipfile, sandbox, **kwargs)
        result = {"asyncId": asyncId, "state": state}
        return result

    # check on a file-based deployment
    @async_retry_expired_session
    async def checkDeployStatus(
        self, asyncId: str, **kwargs: Any
    ) -> Dict[str, None | str | Mapping[str, str]]:
        """Check on the progress of a file-based deployment via Salesforce
        Metadata API.
        Wrapper for SfdcMetaDataApi.check_deploy_status(...).

        Arguments:

        * asyncId: deployment async process ID, returned by Salesforce.deploy()

        Returns status of the deployment the asyncId given.
        """
        (
            state,
            state_detail,
            deployment_detail,
            unit_test_detail,
        ) = await self.mdapi.check_deploy_status(asyncId, **kwargs)
        results = {
            "state": state,
            "state_detail": state_detail,
            "deployment_detail": deployment_detail,
            "unit_test_detail": unit_test_detail,
        }
        return results

    def parse_result_to_json(self, response: httpx.Response) -> Any:
        """ "Parse json from a Response object"""
        return response.json(
            object_pairs_hook=self._object_pairs_hook, parse_float=self._parse_float
        )


class AsyncSFType:
    """An interface to a specific type of SObject"""

    _parse_float = None
    _object_pairs_hook = OrderedDict

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        object_name: str,
        session_id: str,
        sf_instance: str,
        sf_version: Optional[str] = DEFAULT_API_VERSION,
        proxies: Optional[Proxies] = None,
        session: Optional[httpx.AsyncClient] = None,
        session_factory: Optional[Callable[[], httpx.AsyncClient]] = None,
        salesforce: Optional[AsyncSalesforce] = None,
        parse_float: Optional[Callable[[str], Any]] = None,
        object_pairs_hook: Callable[[List[Tuple[Any, Any]]], Any] = OrderedDict,
        request_timeout_seconds: Optional[int] = None,
    ):
        """Initialize the instance with the given parameters.

        Arguments:

        * object_name -- the name of the type of SObject this represents,
                         e.g. `Lead` or `Contact`
        * session_id -- the session ID for authenticating to Salesforce
        * sf_instance -- the domain of the instance of Salesforce to use
        * sf_version -- the version of the Salesforce API to use
        * proxies -- the optional map of scheme to proxy server
        * session -- DEPRECATED. Custom requests session, created in calling code.
                     This enables the use of requests Session features not otherwise
                     exposed by simple_salesforce.
        * session_factory -- Method for generating a new httpx session (AsyncClient).
                    This enables the use of httpx AsyncClient features not
                    otherwise exposed by simple_salesforce.
        * request_timeout_seconds -- Optional request-timeout setting in seconds
                                     (for each request; for more timeout options, use `session`).
        * parse_float -- Function to parse float values with. Is passed along to
                         https://docs.python.org/3/library/json.html#json.load
        * object_pairs_hook -- Function to parse ordered list of pairs in json.
                               To use python 'dict' change it to None or dict.
        """
        if salesforce is None and not session_id:
            raise RuntimeError(
                "The argument session_id or salesforce must be specified to "
                "instanciate SFType."
            )

        self._session_id = session_id
        self.salesforce = salesforce
        self.name = object_name
        self._proxies = proxies
        self._parse_float = parse_float
        self._object_pairs_hook = object_pairs_hook  # type: ignore[assignment]
        self.request_timeout_seconds = request_timeout_seconds

        if session is not None:
            warnings.warn(
                "The session keyword argument for async clients is deprecated"
            )
        if proxies and session:
            logger.warning(
                "Proxies must be defined on custom session object, "
                "ignoring proxies: %s",
                proxies,
            )
        # make sure session_factory is not None
        # override custom session proxies dance
        if (
            not session_factory
            and self.salesforce
            and self.salesforce.session_factory is not None
        ):
            session_factory = self.salesforce.session_factory
        elif not session_factory:
            session_factory = create_session_factory(
                proxies=self._proxies, timeout=self.request_timeout_seconds
            )
        self.session_factory = session_factory

        self.api_usage: MutableMapping[str, Usage | PerAppUsage] = {}
        self.base_url = (
            f"https://{sf_instance}/services/data/v{sf_version}/sobjects"
            f"/{object_name}/"
        )

    @property
    def can_refresh(self) -> bool:
        if self.salesforce:
            return self.salesforce.can_refresh
        return False

    async def refresh_session(self) -> str | None:
        """Pass-through to `AsyncSalesforce` if present"""
        if not self.salesforce:
            return None
        self._session_id = await self.salesforce.refresh_session()
        return self.session_id

    @property
    def session_id(self) -> str | None:
        """Helper to return the session id"""
        if self.salesforce is not None:
            return self.salesforce.session_id
        return self._session_id

    async def metadata(self, headers: Optional[Headers] = None) -> Any:
        """Returns the result of a GET to `.../{object_name}/` as a dict
        decoded from the JSON payload returned by Salesforce.

        Arguments:

        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce("GET", self.base_url, headers=headers)
        return self.parse_result_to_json(result)

    async def describe(self, headers: Optional[Headers] = None) -> Any:
        """Returns the result of a GET to `.../{object_name}/describe` as a
        dict decoded from the JSON payload returned by Salesforce.

        Arguments:

        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce(
            method="GET", url=urljoin(self.base_url, "describe"), headers=headers
        )
        return self.parse_result_to_json(result)

    async def describe_layout(
        self, record_id: str, headers: Optional[Headers] = None
    ) -> Any:
        """Returns the layout of the object

        Returns the result of a GET to
        `.../{object_name}/describe/layouts/<recordid>` as a dict decoded from
        the JSON payload returned by Salesforce.

        Arguments:

        * record_id -- the Id of the SObject to get
        * headers -- a dict with additional request headers.
        """
        custom_url_part = f"describe/layouts/{record_id}"
        result = await self._call_salesforce(
            method="GET", url=urljoin(self.base_url, custom_url_part), headers=headers
        )
        return self.parse_result_to_json(result)

    async def get(
        self, record_id: str, headers: Optional[Headers] = None, **kwargs: Any
    ) -> Any:
        """Returns the result of a GET to `.../{object_name}/{record_id}` as a
        dict decoded from the JSON payload returned by Salesforce.

        Arguments:

        * record_id -- the Id of the SObject to get
        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce(
            method="GET",
            url=urljoin(self.base_url, record_id),
            headers=headers,
            **kwargs,
        )
        return self.parse_result_to_json(result)

    async def get_by_custom_id(
        self, custom_id_field: str, custom_id: str, headers: Optional[Headers] = None
    ) -> Any:
        """Return an ``AsyncSFType`` by custom ID

        Returns the result of a GET to
        `.../{object_name}/{custom_id_field}/{custom_id}` as a dict decoded
        from the JSON payload returned by Salesforce.

        Arguments:

        * custom_id_field -- the API name of a custom field that was defined
                             as an External ID
        * custom_id - the External ID value of the SObject to get
        * headers -- a dict with additional request headers.
        """
        custom_url = urljoin(self.base_url, f"{custom_id_field}/{custom_id}")
        result = await self._call_salesforce(
            method="GET", url=custom_url, headers=headers
        )
        return self.parse_result_to_json(result)

    async def create(
        self, data: Mapping[str, Any], headers: Optional[Headers] = None
    ) -> Any:
        """Creates a new SObject using a POST to `.../{object_name}/`.

        Returns a dict decoded from the JSON payload returned by Salesforce.

        Arguments:

        * data -- a dict of the data to create the SObject from. It will be
                  JSON-encoded before being transmitted.
        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce(
            method="POST", url=self.base_url, content=json.dumps(data), headers=headers
        )
        return self.parse_result_to_json(result)

    async def upsert(
        self,
        record_id: str,
        data: Mapping[str, Any],
        raw_response: bool = False,
        headers: Optional[Headers] = None,
    ) -> int | httpx.Response:
        """Creates or updates an SObject using a PATCH to
        `.../{object_name}/{record_id}`.

        If `raw_response` is false (the default), returns the status code
        returned by Salesforce. Otherwise, return the `requests.Response`
        object.

        Arguments:

        * record_id -- an identifier for the SObject as described in the
                       Salesforce documentation
        * data -- a dict of the data to create or update the SObject from. It
                  will be JSON-encoded before being transmitted.
        * raw_response -- a boolean indicating whether to return the response
                          directly, instead of the status code.
        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce(
            method="PATCH",
            url=urljoin(self.base_url, record_id),
            content=json.dumps(data),
            headers=headers,
        )
        return self._raw_response(result, raw_response)

    async def update(
        self,
        record_id: str,
        data: Mapping[str, Any],
        raw_response: bool = False,
        headers: Optional[Headers] = None,
    ) -> int | httpx.Response:
        """Updates an SObject using a PATCH to
        `.../{object_name}/{record_id}`.

        If `raw_response` is false (the default), returns the status code
        returned by Salesforce. Otherwise, return the `requests.Response`
        object.

        Arguments:

        * record_id -- the Id of the SObject to update
        * data -- a dict of the data to update the SObject from. It will be
                  JSON-encoded before being transmitted.
        * raw_response -- a boolean indicating whether to return the response
                          directly, instead of the status code.
        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce(
            method="PATCH",
            url=urljoin(self.base_url, record_id),
            content=json.dumps(data),
            headers=headers,
        )
        return self._raw_response(result, raw_response)

    async def delete(
        self,
        record_id: str,
        raw_response: bool = False,
        headers: Optional[Headers] = None,
    ) -> int | httpx.Response:
        """Deletes an SObject using a DELETE to
        `.../{object_name}/{record_id}`.

        If `raw_response` is false (the default), returns the status code
        returned by Salesforce. Otherwise, return the `requests.Response`
        object.

        Arguments:

        * record_id -- the Id of the SObject to delete
        * raw_response -- a boolean indicating whether to return the response
                          directly, instead of the status code.
        * headers -- a dict with additional request headers.
        """
        result = await self._call_salesforce(
            method="DELETE", url=urljoin(self.base_url, record_id), headers=headers
        )
        return self._raw_response(result, raw_response)

    async def deleted(
        self, start: datetime, end: datetime, headers: Optional[Headers] = None
    ) -> Any:
        # pylint: disable=line-too-long
        """Gets a list of deleted records

        Use the SObject Get Deleted resource to get a list of deleted records
        for the specified object.
        .../deleted/?start=2013-05-05T00:00:00+00:00&end=2013-05-10T00:00:00
        +00:00

        * start -- start datetime object
        * end -- end datetime object
        * headers -- a dict with additional request headers.
        """
        url = urljoin(
            self.base_url,
            f"deleted/?start={date_to_iso8601(start)}&end={date_to_iso8601(end)}",
        )
        result = await self._call_salesforce(method="GET", url=url, headers=headers)
        return self.parse_result_to_json(result)

    async def updated(
        self, start: datetime, end: datetime, headers: Optional[Headers] = None
    ) -> Any:
        # pylint: disable=line-too-long
        """Gets a list of updated records

        Use the SObject Get Updated resource to get a list of updated
        (modified or added) records for the specified object.

         .../updated/?start=2014-03-20T00:00:00+00:00&end=2014-03-22T00:00:00
         +00:00

        * start -- start datetime object
        * end -- end datetime object
        * headers -- a dict with additional request headers.
        """
        url = urljoin(
            self.base_url,
            f"updated/?start={date_to_iso8601(start)}&end={date_to_iso8601(end)}",
        )
        result = await self._call_salesforce(method="GET", url=url, headers=headers)
        return self.parse_result_to_json(result)

    @async_retry_expired_session
    async def _call_salesforce(
        self, method: str, url: str, **kwargs: Any
    ) -> httpx.Response:
        """Utility method for performing HTTP call to Salesforce.

        Returns a `requests.result` object.
        """
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + (self.session_id or ""),
            "X-PrettyPrint": "1",
        }
        additional_headers = kwargs.pop("headers", dict())
        headers.update(additional_headers or dict())

        async with self.session_factory() as client:
            result = await client.request(method, url, headers=headers, **kwargs)

        if result.status_code >= 300:
            exception_handler(result, self.name)
        sforce_limit_info = result.headers.get("Sforce-Limit-Info")
        if sforce_limit_info:
            self.api_usage = AsyncSalesforce.parse_api_usage(sforce_limit_info)

        return result

    # pylint: disable=no-self-use
    def _raw_response(
        self, response: httpx.Response, body_flag: bool
    ) -> int | httpx.Response:
        """Utility method for processing the response and returning either the
        status code or the response object.

        Returns either an `int` or a `requests.Response` object.
        """
        if not body_flag:
            return response.status_code

        return response

    def parse_result_to_json(self, response: httpx.Response) -> Any:
        """ "Parse json from a Response object"""
        return response.json(
            object_pairs_hook=self._object_pairs_hook, parse_float=self._parse_float
        )

    async def upload_base64(
        self,
        file_path: str,
        base64_field: str = "Body",
        headers: Optional[Headers] = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """Upload base64 encoded file to Salesforce"""
        data = {}
        async with aiofiles.open(file_path, "rb") as fl:
            content = await fl.read()
            body = base64.b64encode(content).decode("utf-8")
        data[base64_field] = body
        result: httpx.Response = await self._call_salesforce(
            method="POST", url=self.base_url, headers=headers, json=data, **kwargs
        )

        return result

    async def update_base64(
        self,
        record_id: str,
        file_path: str,
        base64_field: str = "Body",
        headers: Optional[Headers] = None,
        raw_response: bool = False,
        **kwargs: Any,
    ) -> int | httpx.Response:
        """Updated base64 image from file to Salesforce"""
        data = {}
        async with aiofiles.open(file_path, "rb") as fl:
            content = await fl.read()
            body = base64.b64encode(content).decode("utf-8")
        data[base64_field] = body
        result = await self._call_salesforce(
            method="PATCH",
            url=urljoin(self.base_url, record_id),
            json=data,
            headers=headers,
            **kwargs,
        )

        return self._raw_response(result, raw_response)

    async def get_base64(
        self,
        record_id: str,
        base64_field: str = "Body",
        data: Optional[Mapping[str, Any]] = None,
        headers: Optional[Headers] = None,
        **kwargs: Any,
    ) -> bytes:
        """Returns binary stream of base64 object at specific path.

        Arguments:

        * path: The path of the request
            Example: sobjects/Attachment/ABC123/Body
                     sobjects/ContentVersion/ABC123/VersionData
        """
        result: httpx.Response = await self._call_salesforce(
            method="GET",
            url=urljoin(self.base_url, f"{record_id}/{base64_field}"),
            data=data,
            headers=headers,
            **kwargs,
        )

        return result.content
